package util;

// Java imports
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.text.*;

// UI display text manager
public class Localizer {

    // Instance fields
    private HashMap<Object, Object> controls; // Control mapping
    private HashMap<String, String> messages; // Message dictionary



    ///////////////////////////////////////////////////////////////////////////
    //                               Constants                               //
    ///////////////////////////////////////////////////////////////////////////

    // Class reference for JComboBox<String>
    JComboBox<String> JCOMBOBOX = new JComboBox<String>();



    ///////////////////////////////////////////////////////////////////////////
    //                             Constructors                              //
    ///////////////////////////////////////////////////////////////////////////

    // Default constructor
    public Localizer() {
        controls = new HashMap<Object, Object>();
        messages = new HashMap<String, String>();
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Public Methods                             //
    ///////////////////////////////////////////////////////////////////////////

    // Add a control to the collection
    public boolean add(Object control, Object key) {

        // Error checking
        if (control == null || key == null)
            return false;

        // Control takes a single string
        if (key instanceof String) {
            if (!(
                control instanceof AbstractButton ||
                control instanceof JFrame         ||
                control instanceof JInternalFrame ||
                control instanceof JPanel         || // TitledBorder
                control instanceof JTextComponent
            )) return false;
        }

        // Control takes an array of strings
        else if (key instanceof String[]) {
            if (!(
                JCOMBOBOX.getClass().isAssignableFrom(control.getClass())
            )) return false;
        }

        // Invalid control type
        else return false;

        // Add the control to the collection
        controls.put(control, key);
        update();
        return true;
    }

    // Remove all controls from being managed
    public void clearControls() {
        controls.clear();
    }

    // Configure a dictionary entry
    public String put(String key, String value) {
        String ret = value == null ?
            messages.remove(key) :
            messages.put(key, value)
        ;
        update();
        return ret;
    }

    // Remove a control from the collection
    public boolean remove(Object control) {
        return controls.remove(control) != null;
    }

    // Specify a message dictionary
    public void set(String text) {

        // Pre-processing
        messages.clear();
        if (text == null) {
            update();
            return;
        }

        // Configure working variables
        var chars = text.replaceAll("\\r\\n?", "\n").toCharArray();
        int col   = 1; // Current character on the current line
        int line  = 1; // Current line in the file
        int start = 0; // Position of first character in key or value
        var stack = new Stack<String>(); // Nested key chain
        int state = 0; // Current parsing context

        // Process all characters
        for (int x = 0; x < chars.length; x++, col++) {
            char    c     = chars[x];
            boolean end   = x == chars.length - 1;
            String  pos   = line + ":" + col + ": ";
            boolean white = c == ' ' || c == '\t';

            // Comment
            if (state == -1) {
                if (c != '\n' && !end)
                    continue;
                col   = 1;
                state = 0;
                line++;
            }

            // Pre-key
            else if (state == 0) {

                // Ignore leading whitespace
                if (white)
                    continue;

                // The line is a comment
                if (c == '#') {
                    state = -1;
                    continue;
                }

                // End of input has been reached
                if (end)
                    break;

                // The line is empty
                if (c == '\n') {
                    col = 1;
                    line++;
                    continue;
                }

                // Proceed to key
                start = x;
                state = 1;
            }
                
            // Key
            else if (state == 1) {

                // Any non-whitespace character is valid in a key
                if (!white && c != '\n' && !end)
                    continue;

                // Produce the key as a string
                String key = new String(chars, start, x - start).trim();

                // Close a key group
                if (key.equals("}")) {

                    // Nesting error
                    if (stack.size() == 0) throw new RuntimeException(
                        pos + "Unexpected '}'");

                    // Syntax error
                    outer: for (int y = x; y < chars.length; y++) {
                        char h = chars[y];
                        if (h == '\n') break;
                        if (h != ' ' && h != '\t') throw new RuntimeException(
                            pos + "Newline expected");
                    }

                    // Proceed to key
                    state = 0;
                    stack.pop();
                    continue;
                }

                // Curly braces are not valid in keys
                if (key.contains("{") || key.contains("}"))
                    throw new RuntimeException(
                        line + ": Curly braces are not allowed in keys");

                // Proceed to interim
                stack.push(key);
                state = 2;
            }

            // Post-key, pre-value
            if (state == 2) {

                // Ignore any whitespace between key and value
                if (white)
                    continue;

                // No value is present
                if (c == '\n') throw new RuntimeException(
                    pos + "Unexpected newline");
                if (end) throw new RuntimeException(
                    pos + "Unexpected end of input");

                // Proceed to value
                start = x;
                state = 3;
            }

            // Value
            if (state == 3) {

                // Escape sequence
                if (c == '\\') {

                    // EOF
                    if (x == chars.length - 1)
                        throw new RuntimeException(
                            pos + "Unexpected end of input");

                    // Cannot escape a newline
                    if (chars[x + 1] == '\n')
                        throw new RuntimeException(
                            pos + "Unexpected newline");

                    // Skip the next character
                    x++;
                    continue;
                }

                // The end of the value has not yet been reached
                if (c != '\n' && !end)
                    continue;

                // Produce the value as a string
                if (end)
                    x++;
                String value = new String(chars, start, x - start).trim();
                col   = 1;
                state = 0;
                line++;

                // Open a key group
                if (value.equals("{")) {
                    state = 0;
                    continue;
                }

                // Check for nesting errors
                int depth = 0;
                for (int y = start; y < x; y++) {
                    switch (chars[y]) {
                        case '\\': y++;     continue;
                        case '{' : depth++; continue;
                        case '}' : if (--depth == -1)
                            throw new RuntimeException(
                                line + ":" + (x - start + y + 1) +
                                ": Unexpected '}'"
                            );
                            continue;
                    }
                }
                if (depth != 0) throw new RuntimeException(
                    pos + "'}' expected");

                // Produce the full key as a string
                var pkey = new StringBuilder();
                for (String item : stack)
                    pkey.append((pkey.length() != 0 ? "." : "") + item);
                String key = pkey.toString();

                // Check for duplicate keys
                String lkey = key.toLowerCase();
                if (messages.get(lkey) != null) throw new RuntimeException(
                    (line-1)+ ": Key '" + key + "' has already been defined.");

                // Add the pair to the dictionary
                messages.put(lkey, value);
                stack.pop();
            }

        }

        // Update all controls
        update();
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Private Methods                            //
    ///////////////////////////////////////////////////////////////////////////

    // Process substitutions and escapes on a message for a given key
    private String evaluate(String key) {
        String ret = messages.get(key.toLowerCase());

        // The topmost key does not exist in the dictionary
        if (ret == null)
            return null;

        // Perform all substitutions
        outer: for (;;) {
            var chars = ret.toCharArray();
            int start = 0;

            // Look for the first complete substitution
            for (int x = 0; x < chars.length; x++) {
                char c = chars[x];

                // Escape sequence
                if (c == '\\') {
                    x++;
                    continue;
                }

                // Begin substitution
                if (c == '{')
                    start = x + 1;

                // Substitution has not yet ended
                if (c != '}')
                    continue;

                // Determine the substitution
                       key   = new String(chars, start, x - start);
                String value = messages.get(key.toLowerCase());
                if (value == null)
                    value = "\\{" + key + "\\}";

                // Apply the substitution to the message
                ret =
                    new String(chars, 0, start - 1) +
                    value +
                    new String(chars, x + 1, chars.length - x - 1)
                ;
                continue outer;
            }

            // No further substitutions
            break;
        }

        // Unescape all escape sequences
        var chars  = ret.toCharArray();
        int length = 0;
        for (int x = 0; x < chars.length; x++, length++) {
            char c = chars[x];
            if (c == '\\') switch (c = chars[++x]) {
                case 'n': c = '\n'; break;
                case 't': c = '\t'; break;
            }
            chars[length] = c;
        }
        return new String(chars, 0, length);
    }

    // Update the text for all controls
    private void update() {

        // Process all controls
        for (var control : controls.keySet()) {
            Object   key    = controls.get(control);
            String[] values = null;

            // One string
            if (key instanceof String)
                values = new String[] { evaluate((String) key) };

            // Multiple strings
            else {
                String[] keys = (String[]) key;
                values = new String[keys.length];
                for (int x = 0; x < keys.length; x++)
                    values[x] = evaluate(keys[x]);
            }

            // Update the control's text
            if (control instanceof AbstractButton)
                ((AbstractButton) control).setText (values[0]);
            if (control instanceof JFrame)
                ((JFrame        ) control).setTitle(values[0]);
            if (control instanceof JInternalFrame)
                ((JInternalFrame) control).setTitle(values[0]);
            if (control instanceof JTextComponent)
                ((JTextComponent) control).setText (values[0]);

            // JPanel must be wrapped in a TitledBorder
            if (control instanceof JPanel) {
                var border = ((JPanel) control).getBorder();
                if (border instanceof TitledBorder)
                    ((TitledBorder) border).setTitle(values[0]);
            }

            // Replace the contents of a JComboBox without firing events
            if (JCOMBOBOX.getClass().isAssignableFrom(control.getClass())) {

                // The type is explicitly verified above
                @SuppressWarnings("unchecked")
                var box = (JComboBox<String>) control;

                // Configure working variables
                var action = box.getActionListeners();
                int index  = box.getSelectedIndex();
                var item   = box.getItemListeners();

                // Remove event listeners
                for (var lst : action) box.removeActionListener(lst);
                for (var lst : item  ) box.removeItemListener  (lst);

                // Update contents
                box.setModel(new DefaultComboBoxModel<String>(values));
                box.setSelectedIndex(index);

                // Restore event listeners
                for (var lst : action) box.addActionListener(lst);
                for (var lst : item  ) box.addItemListener  (lst);
            }

        }

    }

}
